package com.capitalware.ufm.actions;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Vector;

import org.jdom.JDOMException;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.ChannelSftp.LsEntry;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.SFtpObject;
import com.capitalware.ufm.ftp.objects.*;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.internal.messages.UpdateStatus;
import com.capitalware.ufm.objects.ConnectionObject;
import com.capitalware.ufm.xml.SFtpXml;

/**
 * This class will handle the SFTP action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class SFtpAction extends UFMAction
{
   private     SFtpObject     sfo;
   private     SFtpXml        sfxh;
   private     InternalMessageQueue   iq = null;
   private     String         fileSeparator = "/"; 

   /**
    * The constructor
    * @param sfo
    */
   public SFtpAction(SFtpObject sfo, InternalMessageQueue iq)
   {
      super(sfo.actionName);
      this.sfo = sfo;
      this.iq = iq;
   }

   /**
    * Perform the work.
    * @return exitCode
    */
   public int doWork()
   {
      int                     exitCode = 0;
      ConnectionObject        co = null;
      ArrayList<FtpCmdObject> cmdsAL = null;
      JSch                    jsch = new JSch();
      Session                 session = null;
      Channel                 channel = null;
      ChannelSftp             sftpChannel = null;
      
      sfo.printValues();
      
      sfxh = new SFtpXml();
      try
      {
         sfxh.readXMLFile(sfo.xmlFile);
         
         co = sfxh.getFtpConnection();
         cmdsAL = sfxh.getSFtpCmds();

         /* Dump out the values */
         co.printValues("SFTP");
         printValues(cmdsAL);

         /* Check/validate the values */
         co.checkObject("SFTP");
         checkObject(cmdsAL);

         /* */
         java.util.Properties config = new java.util.Properties();
         config.put("StrictHostKeyChecking", "no");

         /* connect to the SFTP server. */
         session = jsch.getSession(co.userID, co.hostname, co.port);
         
         session.setConfig(config);
         session.setPassword(co.password);
         session.connect();
         
         channel = session.openChannel("sftp");
         channel.connect();
         sftpChannel = (ChannelSftp) channel;
         
         String curDir = sftpChannel.pwd();
         UFM.logger.debug("Current directory is '" + curDir+"'");
         
         if (curDir.indexOf('\\') != -1) /* windows SFTP server? */
            fileSeparator = "\\";
         
         /* Process SFTP commands. */
         exitCode = processSftpCommands(session, channel, sftpChannel, cmdsAL);
      }
      catch (JDOMException e)
      {
         UFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (JSchException e)
      {
         UFM.logger.fatal("JSchException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      catch (SftpException e)
      {
         UFM.logger.fatal("SftpException: " + e.getLocalizedMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 8;
      }
      finally
      {
         if (sftpChannel != null)
         {
            sftpChannel.quit();
            UFM.logger.info("Issued quit command.");
         }
         
         if (session != null)
         {
            session.disconnect();
            UFM.logger.info("Disconnected from " + co.hostname);
         }

         if (!sfo.onErrorFail)
            exitCode = 0; /* reset back to zero */
      }
      
      return exitCode;
   }
   
   
   /**
    * Print out SFTP command objects, if any.
    * @param cmdsAL
    */
   private void printValues(ArrayList<FtpCmdObject> cmdsAL)
   {
      if (cmdsAL != null)
      {
         UFM.logger.info("     SFTP commands to be processed:");
         for (int i=0; i < cmdsAL.size(); i++)
         {
            UFM.logger.info("       SFTP Command ["+(i+1)+"]: " + ((FtpCmdObject)cmdsAL.get(i)).cmdName);
            ((FtpCmdObject)cmdsAL.get(i)).printValues();
         }
      }
      
   }
   
   /**
    * Issue checkObject for each SFTP command object
    * @param cmdsAL
    */
   private void checkObject(ArrayList<FtpCmdObject> cmdsAL)
   {
      if (cmdsAL != null)
      {
         for (int i=0; i < cmdsAL.size(); i++)
         {
            ((FtpCmdObject)cmdsAL.get(i)).performTokenization(sfo.paClass.actionProperties);
            ((FtpCmdObject)cmdsAL.get(i)).checkObject();
         }
      }
   }
   
   /**
    * Loop through SFTP commands and process them.
    * @param session
    * @param channel
    * @param sftpChannel
    * @param cmdsAL
    * @return exitCode
    * @throws IOException
    * @throws SftpException 
    * @throws JSchException 
    */
   private int processSftpCommands(Session session, Channel channel, ChannelSftp sftpChannel, ArrayList<FtpCmdObject> cmdsAL) throws IOException, SftpException, JSchException
   {
      SimpleDateFormat  dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int               exitCode = 0;
      FtpCmdObject      sftpco = null;
      
      /**
       * Process SFTP commands.
       */
      if (cmdsAL != null)
      {
         for (int i=0; i < cmdsAL.size(); i++)
         {
            sftpco = cmdsAL.get(i);
            if (sftpco instanceof GetObject)
            {
               exitCode = processGetCommand(sftpChannel, (GetObject)sftpco);
            }
            else if (sftpco instanceof PutObject)
            {
               exitCode = processPutCommand( sftpChannel, (PutObject)sftpco);
            }
            else if (sftpco instanceof CdObject)
            {
               CdObject co = (CdObject)sftpco;
               sftpChannel.cd(co.targetDir);
               UFM.logger.info("Changed directory to " + co.targetDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=cd "+co.targetDir+"\n"));
            }
            else if (sftpco instanceof MkDirObject)
            {
               MkDirObject mdo = (MkDirObject)sftpco;
               sftpChannel.mkdir(mdo.remoteDir);
               UFM.logger.info("Created directory: " + mdo.remoteDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=mkdir "+mdo.remoteDir+"\n"));
            }
            else if (sftpco instanceof RmDirObject)
            {
               RmDirObject rdo = (RmDirObject)sftpco;
               sftpChannel.rmdir(rdo.remoteDir);
               UFM.logger.info("Deleted directory: " + rdo.remoteDir);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=rmdir "+rdo.remoteDir+"\n"));
            }
            else if (sftpco instanceof DeleteObject)
            {
               DeleteObject dlo = (DeleteObject)sftpco;
               String deletedFile = IO.apendFileToPath(dlo.remoteDir, dlo.remoteFile, "/");
               sftpChannel.rm(deletedFile);
               UFM.logger.info("Deleted file: " + deletedFile);

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=del "+deletedFile+"\n"));
            }
         }
      }
      
      return exitCode;
   }
   
   /**
    * Process the get command
    * @param sftpChannel
    * @param go
    * @return exitCode
    * @throws IOException
    * @throws SftpException 
    */
   private int processGetCommand(ChannelSftp sftpChannel, GetObject go) throws IOException, SftpException
   {
      SimpleDateFormat  dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int               exitCode = 0;
      String            remoteFile; 
      String            localFile;
      LsEntry           lsEntry;

      if (go.remoteFilename.indexOf('*') != -1)
      {
         @SuppressWarnings("unchecked")
         Vector<LsEntry> files = sftpChannel.ls(go.remoteFilename);
         
          for (int i=0; i<files.size(); i++)
          {
              lsEntry = files.get(i);

              if (!lsEntry.getFilename().equals(".") && !lsEntry.getFilename().equals("..")) 
              {
                 localFile = IO.apendFileToPath(go.localDir, lsEntry.getFilename());

                  //Get the file and write it to our local file system
                  File f = new File(localFile);
                  sftpChannel.get(lsEntry.getFilename(), new FileOutputStream(f));
                  UFM.logger.info("Retrieved file from " + lsEntry.getFilename() + " and stored as " + localFile);
                  go.fileCount++;

                  if (iq != null)
                     iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=get "+lsEntry.getFilename()+"\n"));
              }
          }
      }
      else
      {
         remoteFile = IO.apendFileToPath(go.remoteDir, go.remoteFilename, fileSeparator);
         localFile = IO.apendFileToPath(go.localDir, (go.localFilename == null?go.remoteFilename:go.localFilename));
         
         File f = new File(localFile);
         sftpChannel.get(remoteFile, new FileOutputStream(f));
         UFM.logger.info("Retrieved file from " + remoteFile + " and stored as " + localFile);
         go.fileCount++;

         if (iq != null)
            iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=get "+remoteFile+"\n"));
      }
      
      return exitCode;
   }
   
   /**
    * Process the put command
    * @param sftpChannel
    * @param po
    * @return exitCode
    * @throws IOException
    * @throws SftpException 
    */
   private int processPutCommand(ChannelSftp sftpChannel, PutObject po) throws IOException, SftpException
   {
      SimpleDateFormat  dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int               exitCode = 0;
      String            remoteFile; 
      String            localFile;

      if (po.localFilename.indexOf('*') != -1)
      {
         File[] fileList;
         
         try
         {
            if (po.localDir != null)
               fileList = IO.getFilteredFileList(IO.apendFileToPath(po.localDir, po.localFilename));
            else
               fileList = IO.getFilteredFileList(po.localFilename);
            
            for (int i=0; i < fileList.length; i++)
            {
               String toFilename = fileList[i].getName();

               remoteFile = IO.apendFileToPath(po.remoteDir, toFilename, fileSeparator);
               remoteFile = Tokens.replace("put", remoteFile, toFilename, i+1, sfo.paClass.actionProperties);
               
               localFile = IO.apendFileToPath(po.localDir, toFilename);
               sftpChannel.put(localFile, remoteFile);
               UFM.logger.info("Stored file " + localFile + " to " + remoteFile);
               po.fileCount++;

               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=put "+remoteFile+"\n"));
            }
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            UFM.globalErrorText = e.getLocalizedMessage();
            exitCode = 8;
         }
      }
      else
      {
//         ChannelExec ce = (ChannelExec)sftpChannel;
         
         remoteFile = IO.apendFileToPath(po.remoteDir, (po.remoteFilename == null?po.localFilename:po.remoteFilename), fileSeparator);
         localFile = IO.apendFileToPath(po.localDir, po.localFilename);
         
         sftpChannel.put(localFile, remoteFile);
         UFM.logger.info("Stored file " + localFile + " to " + remoteFile);
         po.fileCount++;

         if (iq != null)
            iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+sfo.actionName+", "+sfo.labelName+", SFTP_Command=put "+remoteFile+"\n"));
      }
      
      return exitCode;
   }
}
