package com.capitalware.ufm.handlers;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.Vector;

/**
 * This is a set of string utilities,
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class StringUtilities
{
   private static final String HEX = "0123456789ABCDEF";

   /**
    * Make sure we don't call it.
    */
   private StringUtilities()
   {}

   public static String replaceAll(String buffer, String find, String replace)
   {
      String h, s = buffer;
      int i = s.indexOf(find);
      while (i >= 0) 
      {
         h = s.substring(0, i) + replace;
         s = h + s.substring(i + find.length());
         i = s.indexOf(find, h.length());
      }

      return s;
   }

	/**
    * This method will pad the the left side of the string argument with the padding string until
    * the length is reached. The padding argument is only one character, it will be shortened if
    * length > 1
    *
	 * @param aString java.lang.String
	 * @param padding java.lang.String
	 * @param length int
    * @return java.lang.String
	 */
   public static String padLeft(String aString, String padding, int length)
   {
      if (aString == null || padding == null || length < 1 || aString.length() >= length)
         return aString;

      if (padding.length() > 1)
         padding = padding.substring(0, 1);

      int padLen = length - aString.length();

      StringBuffer buf = new StringBuffer();

      for (int i=0; i < padLen; i++)
      {
         buf.append(padding);
      }

      buf.append(aString);

      return buf.toString();
   }

	/**
    * This method will pad the the right side of the string argument with the padding string until
    * the length is reached. The padding argument is only one character, it will be shortened if
    * length > 1
	 *
	 * @param aString java.lang.String
	 * @param padding java.lang.String
	 * @param length int
    * @return java.lang.String
	 */
   public static String padRight(String aString, String padding, int length)
   {
      if (aString == null || padding == null || length < 1 || aString.length() >= length)
         return aString;

      if (padding.length() > 1)
         padding = padding.substring(0, 1);

      StringBuffer buf = new StringBuffer();
      buf.append(aString);

      while (buf.length() < length)
      {
         buf.append(padding);
      }

      return buf.toString();
   }

	/**
	 * Packs the given array, removing any empty or null elements.
	 * @param limits A String array that may contain empty or null elements. Returns null
	 * if the original array is null or no non-empty elements exist in the array.
	 */
	public static String[] removeEmptyStrings(String[] anArray)
	{
		String[] tempArray = null;
		String[] newArray = null;
		int j = 0;
		try
		{
			tempArray = new String[anArray.length];
			for (int i = 0; i < anArray.length; i++)
			{
				if (anArray[i] != null)
				{
					if (anArray[i].length() > 0)
					{
						tempArray[j++] = anArray[i];
					}
				}
			}
		}
		catch (NullPointerException e)
		{
			return null;
		}
		if (j == anArray.length)
		{
			return anArray;
		}
		newArray = new String[j];
		System.arraycopy(tempArray, 0, newArray, 0, j);
		return newArray;
	}

	/**
    * This method strips out any non-numeric characters from the string argument.
	 * @param aString java.lang.String
    * @return java.lang.String
	 */
   public static String toDigitsOnly(String input)
   {
      if (input == null)
         return input;

      if (input.length() < 1)
         return input;

      String aString = input.trim();
      int j = 0;
      for (int i = 0; i < aString.length(); i++)
      {
         char c = aString.charAt(i);
         if (Character.isDigit(c))
         {
            j++;
         }
      }

      if (j == 0)
         return aString;

      char digits[] = new char[j];
      int k = 0;
      for (int i = 0; i < aString.length(); i++)
      {
         char c = aString.charAt(i);
         if (Character.isDigit(c))
         {
            digits[k] = c;
            k++;
            if (k >= j)
               break;
         }
      }

      return new String(digits);
   }

   /**
    *
    * @param s1
    * @param len
    * @return
    */
   public static StringBuffer getFixedSB(String s1, int len)
   {
      int ll = s1.length();
      StringBuffer sb = new StringBuffer(len);
      sb.append(s1);

      for (int i=ll; i < len; i++)
      {
         sb.insert(i, '\0');
      }
      sb.setLength(len);


      return sb;
   }

   /**
    * Convert a StringBuffer to a String Array.
    *
    * @param data
    * @return String[]
    * @throws IOException
    */
   public static String[] toArrayOfString(StringBuffer data) throws IOException
   {
      DataInputStream dis = new DataInputStream(new ByteArrayInputStream((data.toString()).getBytes()));

      Vector<String> s = new Vector<String>();
      for (;;)
      {
         String line;
         line = dis.readLine();
         if (line == null)
            break;
         s.addElement(line);
      }
      String[] a = new String[s.size()];
      s.copyInto(a);
      return a;
   }
   
   /**
    * Convert a byte[] to a String Array.
    *
    * @param data
    * @return String[]
    * @throws IOException
    */
   public static String[] toArrayOfString(byte[] data) throws IOException
   {
      DataInputStream dis = new DataInputStream(new ByteArrayInputStream(data));

      Vector<String> v = new Vector<String>();
      for (;;)
      {
         String line;
         line = dis.readLine();
         if (line == null)
            break;
         v.addElement(line);
      }
      String[] a = new String[v.size()];
      v.copyInto(a);
      return a;
   }
   
   /**
    * Convert a String[] to a String
    *
    * @param data
    * @return String[]
    * @throws IOException
    */
   public static String toStringFromArrayOfString(String[] data)
   {
      StringBuffer out = new StringBuffer();
      int len = data.length;
      
      for (int i=0; i < len; i++)
         out.append(data[i]+";");
      
      return out.toString();
   }
   
   /**
    * Convert a String to an int.
    *
    * @param data
    * @param defaultValue
    * @return
    */
   public static int stringToInt(String data, int defaultValue)
   {
      int x = defaultValue;

      try
      {
         x = Integer.parseInt(data);
      }
      catch (Exception e)
      {
         x = defaultValue;
      }

      return x;
   }

   /**
    * Convert a byte array to a hex string representation
    * @param byteArray
    * @return string
    */
   public static String byteArrayToHexString(byte[] byteArray)
   {
      int len = byteArray.length;
      StringBuffer hexStr = new StringBuffer();

      for (int j=0; j < len; j++)
         hexStr.append(byteToHex((char)byteArray[j]));

      return hexStr.toString();
   }

   /**
    * Convert a byte to its hexadecimal value.
    * @param val
    * @return
    */
   private static String byteToHex(char val)
   {
     int hi = (val & 0xF0) >> 4;
     int lo = (val & 0x0F);
     return "" + HEX.charAt(hi) + HEX.charAt(lo);
   }
}
