package com.capitalware.ufm.mq;

import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.Date;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.internal.messages.*;
import com.capitalware.ufm.mq.objects.MQObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.UFMException;

import com.ibm.mq.*;
import com.ibm.mq.constants.CMQC;

/**
 * This class will handle connection to the queue manager, opening the status queue
 * and getting/putting messages to/from the status queue. 
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class MQStatusUpdate extends MyMQ
{
   private static final int TENTHS_OF_A_SECOND_IN_A_DAY = 864000; // 24 * 60 * 60 * 10;
   
   private MQQueue      statusQ = null;
   private boolean      loopAgain = true;
   private boolean      running = true;
   private int          timeToLive = CWConstants.TIME_TO_LIVE_DEFAULT; // number of days
   private InternalMessageQueue iq = null;       // internal queue
   private byte[]       currentCorrelID = null;
   private String       headerMessage = null;
   private long         lastMessageSize = 0;
   
   /**
    * The constructor
    */
   public MQStatusUpdate(MQObject qm)
   {
   	super(qm);
   }

   /** 
    * Set the internal queue object
    * @param iq
    */
   public void setIQ(InternalMessageQueue iq)
   {
      this.iq = iq;
   }

   /** 
    * Set the status message's time to live
    * @param timeToLive
    */
   public void setTimeToLive(int timeToLive)
   {
      this.timeToLive = timeToLive;
   }

   /**
    * Exit loop processing
    */
   public synchronized void  stopIt()
   {
      loopAgain = false;
   }

   /**
    * Is the thread still running?
    * @return
    */
   public boolean isRunning()
   {
      return running;
   }
   
   /**
    * Run (start) this puppy.
    */
   public void run()
   {
      createCorrelID();
      
      try
      {
         connect();
         
         statusQ = openInOutQ(mqo.qObject);
      }
      catch (MQException e)
      {
         UFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
         UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (InvalidAlgorithmParameterException e)
      {
         UFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (NoSuchAlgorithmException e)
      {
         UFM.logger.fatal("NoSuchAlgorithmException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (ClassNotFoundException e)
      {
         UFM.logger.fatal("ClassNotFoundException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (InstantiationException e)
      {
         UFM.logger.fatal("InstantiationException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (IllegalAccessException e)
      {
         UFM.logger.fatal("IllegalAccessException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (UFMException e)
      {
         UFM.logger.fatal("UFMException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      
      doWork();
      
      // Done. Close the queue.
      closeQ(statusQ);
      // Disconnect from the queue manager.
      disconnect();
      
      running = false;
   }

   /**
    * Main loop for processing the internal queue
    */
   private void doWork()
   {
      Object o;
      
      while (loopAgain)
      {
         o = iq.getMessage();
         
         if (o == null)
         {
            // Nothing on the queue!
            try  // sleep.
            {
               Thread.sleep(250); // 1/4 second
            }
            catch(Exception ee) 
            {}
         }
         else if (o instanceof UpdateStatus)
         {
            String status = ((UpdateStatus)o).getStatus();
            String msg = ((UpdateStatus)o).getMsg();
//            
//            UFM.logger.debug("MQStatusUpdate: status='"+status+"'");
//            UFM.logger.debug("MQStatusUpdate: msg='"+msg+"'");

            if (statusQ != null)
            {
               /* Make sure the status message is under the 4MB limit
                * otherwise create a new message. */
               if (lastMessageSize > CWConstants.MAX_GROUP_MSG_SIZE)
               {
                  createCorrelID();
                  putStatusMessage(headerMessage+msg);
               }
               else
               {
                  updateStatusMessage(status, msg);
               }
            }
         }
         else if (o instanceof NewStatus)
         {
            headerMessage = ((NewStatus)o).toString();
//            UFM.logger.debug("MQStatusUpdate: headerMessage='"+headerMessage+"'");
            if (statusQ != null)
               putStatusMessage(headerMessage);
            
         }
         else if (o instanceof Quit)
         {
            loopAgain = false;
            UFM.logger.info("MQStatusUpdate: quit message received.");
         }
      }
   }
   
   /**
    * Create a unique CorrelID.
    */
   private void createCorrelID()
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyyMMddHHmmssSSS");
      String prefix = Thread.currentThread().hashCode() + "       ";

      prefix = prefix.substring(0, 7) + dateTimeFormat.format(new Date());

      currentCorrelID = prefix.getBytes();
   }
   
   /**
    * Put an initial status message on the queue.
    * @param msg
    */
   private void putStatusMessage(String msg)
   {
      MQPutMessageOptions pmo = new MQPutMessageOptions();
      pmo.options = CMQC.MQPMO_FAIL_IF_QUIESCING;
      
      MQMessage sendmsg = new MQMessage();

      sendmsg.messageId    = CMQC.MQMI_NONE;
      sendmsg.correlationId= currentCorrelID;
      sendmsg.feedback     = CWConstants.UFM_STATUS_MSG;
      sendmsg.format       = CMQC.MQFMT_STRING;
      sendmsg.messageType  = CMQC.MQMT_DATAGRAM;
      sendmsg.expiry       = timeToLive * TENTHS_OF_A_SECOND_IN_A_DAY; // expiry is in tenths of a second

      lastMessageSize = msg.length();
      
      try
      {
         sendmsg.write(msg.getBytes());

         doPut(pmo, sendmsg);
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
   }
   
   /**
    * Put the message on the queue.
    * @param pmo
    * @param sendmsg
    */
   private void doPut(MQPutMessageOptions pmo, MQMessage sendmsg)
   {
      try
      {
         // put the message on the queue.
         statusQ.put(sendmsg, pmo);

         UFM.logger.info("Put status message on " + mqo.qObject);
      }
      catch (MQException e)
      {
         UFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
         UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         reconnect();
         doPut(pmo, sendmsg);
      }
   }
   
   /**
    * Get the status message from the queue, update it and put it back.
    * Yes, we are treating the queue as a database!!
    * @param msg
    */
   private void updateStatusMessage(String status, String msg)
   {
      byte[] b;
      int x =0;
      MQGetMessageOptions gmo = new MQGetMessageOptions();
      gmo.matchOptions = CMQC.MQMO_MATCH_CORREL_ID;
      gmo.options = CMQC.MQGMO_CONVERT + CMQC.MQGMO_NO_WAIT + CMQC.MQGMO_FAIL_IF_QUIESCING;

      MQMessage receiveMsg = new MQMessage();
      receiveMsg.messageId     = CMQC.MQMI_NONE;
      receiveMsg.correlationId = currentCorrelID;

      try
      {
         doGet(gmo, receiveMsg);
         
         b = new byte[receiveMsg.getMessageLength()];
         receiveMsg.readFully(b);
         String curMsg = new String(b);

         UFM.logger.info("Retrieved status message from " + mqo.qObject);

         /*
          * Ok we have the old status message
          */
         if ((receiveMsg.feedback & CWConstants.UFM_STATUS_MSG) == CWConstants.UFM_STATUS_MSG)
         {
            for (int i=0; i < 5;i++)
            {
               if ( (x = curMsg.indexOf(',', x)) != -1)
               {
                  x++;
               }
               else
                  break;
            }
            
            if ( x != -1)
            {
               String prefix = curMsg.substring(0, x);
               String suffix = curMsg.substring(x+2);
               
               curMsg = prefix + " " + status + suffix;
            }
         }
         
         curMsg += msg;
         putStatusMessage(curMsg);
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
   }
   
   /**
    * Get the message from the queue.
    * @param pmo
    * @param sendmsg
    */
   private MQMessage doGet(MQGetMessageOptions gmo, MQMessage receiveMsg)
   {
      try
      {
         statusQ.get(receiveMsg, gmo);
      }
      catch (MQException e)
      {
         if ( ( (e.completionCode == CMQC.MQCC_WARNING) && (e.reasonCode == CMQC.MQRC_FORMAT_ERROR) )           ||
              ( (e.completionCode == CMQC.MQCC_WARNING) && (e.reasonCode == CMQC.MQRC_TRUNCATED_MSG_ACCEPTED) ) ||
              ( (e.completionCode == CMQC.MQCC_WARNING) && (e.reasonCode == CMQC.MQRC_DBCS_ERROR) )             )
         {
            // ok partial message of some sort
         }
         else if ( (e.completionCode == CMQC.MQCC_FAILED) && (e.reasonCode == CMQC.MQRC_NO_MSG_AVAILABLE) )
         {
            // no message - return an empty message
         }
         else
         {
            UFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
            UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
            reconnect();
            doGet(gmo, receiveMsg);
         }
      }
      
      return receiveMsg;
   }
   
   /**
    * Attempt to reconnect to the queue manager.
    */
   private void reconnect()
   {
      boolean more = true;

      UFM.logger.debug("MQStatusUpdate(): reconnect()");

      // Done. Close the queue.
      if (statusQ != null)
      {
         closeQ(statusQ);
         statusQ = null;
      }
      // Disconnect from the queue manager.
      disconnect();

      while (more)
      {
         // sleep for 30 seconds - 1 second at a time just in case the service
         // needs to be shutdown while we are sleeping
         for (int i=0; i < 30; i++)
         {
            try
            {
               Thread.sleep(1000); // time in milliseconds
            }
            catch (InterruptedException ie) {}

            if (!loopAgain)  // is this thread suppose to stop?
            {
               more = false;
               break;
            }
         }

         try
         {
            connect();

            statusQ = openInOutQ(mqo.qObject);

            if (statusQ == null)
               more = true;
            else
               more = false;
         }
         catch (MQException e)
         {
            UFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
            UFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         }
         catch (Exception e)
         {
            UFM.logger.fatal("Exception: " + e.getMessage());
         }
         finally
         {
            // if more is true then the open may have failed - hence disconnect and try again.
            if (more)
            {
               if (statusQ != null)
               {
                  closeQ(statusQ);
                  statusQ = null;
               }
               
               disconnect();
            }
         }
      }
   }
}
